"""Utils for process manager."""

import asyncio
import inspect
import platform
import socket
from copy import copy
from ipaddress import ip_address
from typing import List, Type

import log

from .cmd_parser import ManagerCMDParser
from .config_loader import ConfigLoader
from .manager import Manager
from .process import Process, SanicProcess


def createManager(
    processes: List[Process],
    configClass: Type["BasePythonServiceSettingsClass"],
    ctxClass: Type["LunaAppCtx"] | None = None,
    port: int | None = None,
) -> Manager:
    """
    Fabric for creation of the manager.

    Args:
        processes: List of processes to manage
        configClass: configuraiton class
        ctxClass: application context class
    """
    if platform.system() != "Windows":
        import uvloop  # pylint: disable-msg=import-outside-toplevel

        asyncio.set_event_loop(uvloop.new_event_loop())
    argParser = ManagerCMDParser(tagableSettings=configClass.getServiceSettingNames(), defaultPort=port)
    cmdArgs = argParser.parse()

    # Leaving "ConfigLoader" and its initialization outside of the manager to stop
    # propagating its arguments (which are constant) through frames.
    configLoader = ConfigLoader(
        configClass, cmdArgs.config, configuratorAddress=cmdArgs.configuratorAddress, tags=cmdArgs.tags
    )
    initialConfig = configLoader.load()

    # Create logger as early as possible, so we'll be able to correctly log occurring exceptions.
    log.initiate(
        appName=initialConfig.serviceName,
        logLevel=initialConfig.logger.logLevel,
        logTime=initialConfig.logger.logTime,
        folderForLog=initialConfig.logger.folderWithLogs,
        maxSize=initialConfig.logger.maxLogFileSize,
        logNameSuffix=cmdArgs.logSuffix,
        logToStdOut=initialConfig.logger.logToStdOut,
        logToFile=initialConfig.logger.logToFile,
        multilineStackTrace=initialConfig.logger.multilineStackTrace,
        formatter=initialConfig.logger.format,
    )

    # Find a sanic process
    for process in processes:
        if isinstance(process, SanicProcess):
            sanicProcess = process
            break
    else:
        raise RuntimeError("At least 1 sanic process should be used.")

    sig = inspect.signature(sanicProcess.target)
    sock = None
    if "sock" in sig.parameters:
        if port is None:
            raise RuntimeError("Port must be specified for a shared socket")
        sock = bindSocket("0.0.0.0", cmdArgs.port)
        sanicProcess.sock = sock

    # spawn workers
    for i in range(1, cmdArgs.workers):
        worker = copy(sanicProcess)
        worker.name += f" {i+1}"
        processes.append(worker)

    manager = Manager(processes, initialConfig=initialConfig, ctxClass=ctxClass, sock=sock)

    if cmdArgs.configReload:
        manager.enableConfigReload(configLoader, period=cmdArgs.configPullingPeriod)

    return manager


def bindSocket(host: str, port: int, backlog=100):
    """
    Create TCP server socket.

    Args:
        host: IPv4, IPv6 or hostname may be specified
        port: TCP port number
        backlog: Maximum number of connections to queue

    Returns socket.socket object
    """
    location = (host, port)
    try:  # IP address: family must be specified for IPv6 at least
        ip = ip_address(host)
        host = str(ip)
        sock = socket.socket(socket.AF_INET6 if ip.version == 6 else socket.AF_INET)
    except ValueError:  # Hostname, may become AF_INET or AF_INET6
        sock = socket.socket()
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    if platform.system() != "Windows":
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    sock.bind(location)
    sock.listen(backlog)
    sock.set_inheritable(True)
    return sock
