"""Simple pool."""

import asyncio
from contextlib import asynccontextmanager
from typing import AsyncContextManager


class SimplePool[CL]:
    """
    Simple pool.

    Attributes:
        queue: asyncio.Queue for items rotation
        items: List of items
    """

    def __init__(self):
        self.queue = asyncio.Queue()
        self.items = []

    def addItem(self, item: CL):
        """Add item to the pool."""
        self.items.append(item)
        self.queue.put_nowait(item)

    @asynccontextmanager
    async def getItem(self) -> AsyncContextManager[CL]:
        """Asynchronous context manager."""
        item = await self.queue.get()
        try:
            yield item
        finally:
            self.queue.put_nowait(item)
